@@ -58,6 +58,7 @@ import GHCi.RemoteTypes
 import GHC.Iface.Load
 import GHCi.Message (ConInfoTable(..), LoadedDLL)
 
+import GHC.ByteCode.Breakpoints
 import GHC.ByteCode.Linker
 import GHC.ByteCode.Asm
 import GHC.ByteCode.Types
@@ -1711,8 +1712,8 @@ allocateCCS interp ce mbss
               let count = maybe 0 ((+1) . fst) $ IM.lookupMax imodBreaks_breakInfo
               let ccs = IM.map
                     (\info ->
-                      fromMaybe (toRemotePtr nullPtr)
-                        (M.lookup (cgb_tick_id info) ccss)
+                        fromMaybe (toRemotePtr nullPtr)
+                          (M.lookup (either internalBreakLoc id (cgb_tick_id info)) ccss)
                     )
                     imodBreaks_breakInfo
               assertPpr (count == length ccs)
