@@ -1,4 +1,3 @@
-{-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE CPP #-}
 
 import Control.Exception
@@ -27,11 +26,8 @@ import Test.Cabal.Prelude
 main = setupAndCabalTest . recordMode DoNotRecord $ do
   -- Foreign libraries don't work with GHC 7.6 and earlier
   skipUnlessGhcVersion ">= 7.8"
-  win <- isWindows
   ghc94 <- isGhcVersion ">= 9.4.1"
-  ghc844 <- isGhcVersion "== 8.4.4"
-  expectBrokenIf (ghc844 || (win && ghc94)) 8451 $
-    withPackageDb $ do
+  withPackageDb $ do
         setup_install []
         setup "copy" [] -- regression test #4156
         dist_dir <- fmap testDistDir getTestEnv
