@@ -1,15 +1,12 @@
-#if __GLASGOW_HASKELL__ < 802
-{-# OPTIONS_GHC -Wno-incomplete-patterns -Wno-incomplete-uni-patterns #-}
-#endif
-module Test.Calendar.CalendarProps
-    ( testCalendarProps
-    ) where
+module Test.Calendar.CalendarProps (
+    testCalendarProps,
+) where
 
 import Data.Time.Calendar.Month
 import Data.Time.Calendar.Quarter
-import Test.TestUtil
-import Test.Tasty
 import Test.Arbitrary ()
+import Test.Tasty
+import Test.TestUtil
 
 testYearMonth :: TestTree
 testYearMonth = nameTest "YearMonth" $ \m -> case m of
@@ -24,4 +21,4 @@ testYearQuarter = nameTest "YearQuarter" $ \q -> case q of
     YearQuarter y qy -> q == YearQuarter y qy
 
 testCalendarProps :: TestTree
-testCalendarProps = nameTest "calender-props" [testYearMonth,testMonthDay,testYearQuarter]
+testCalendarProps = nameTest "calendar-props" [testYearMonth, testMonthDay, testYearQuarter]
